use nom::branch::alt;
use nom::bytes::complete::is_not;
use nom::bytes::complete::tag_no_case;
use nom::character::complete::line_ending;
use nom::character::complete::space0;
use nom::character::complete::space1;
use nom::combinator::consumed;
use nom::combinator::eof;
use nom::combinator::opt;
use nom::combinator::recognize;
use nom::combinator::verify;
use nom::multi::many_till;
use nom::sequence::tuple;

use super::org_source::OrgSource;
use crate::context::parser_with_context;
use crate::context::ContextElement;
use crate::context::ContextMatcher;
use crate::context::ExitClass;
use crate::context::ExitMatcherNode;
use crate::context::RefContext;
use crate::error::Res;
use crate::parser::object_parser::standard_set_object;
use crate::parser::util::blank_line;
use crate::parser::util::exit_matcher_parser;
use crate::parser::util::get_consumed;
use crate::parser::util::start_of_line;
use crate::parser::util::text_until_exit;
use crate::types::CommentBlock;
use crate::types::ExampleBlock;
use crate::types::ExportBlock;
use crate::types::Object;
use crate::types::PlainText;
use crate::types::SrcBlock;
use crate::types::VerseBlock;

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
pub(crate) fn verse_block<'b, 'g, 'r, 's>(
    context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
) -> Res<OrgSource<'s>, VerseBlock<'s>> {
    let (remaining, name) = lesser_block_begin("verse")(context, input)?;
    let (remaining, parameters) = opt(tuple((space1, data)))(remaining)?;
    let (remaining, _nl) = recognize(tuple((space0, line_ending)))(remaining)?;
    let lesser_block_end_specialized = lesser_block_end("verse");
    let contexts = [
        ContextElement::ConsumeTrailingWhitespace(true),
        ContextElement::Context("lesser block"),
        ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &lesser_block_end_specialized,
        }),
    ];
    let parser_context = context.with_additional_node(&contexts[0]);
    let parser_context = parser_context.with_additional_node(&contexts[1]);
    let parser_context = parser_context.with_additional_node(&contexts[2]);
    let parameters = match parameters {
        Some((_ws, parameters)) => Some(parameters),
        None => None,
    };

    let object_matcher = parser_with_context!(standard_set_object)(&parser_context);
    let exit_matcher = parser_with_context!(exit_matcher_parser)(&parser_context);
    // Check for a completely empty block
    let (remaining, children) = match consumed(many_till(blank_line, exit_matcher))(remaining) {
        Ok((remaining, (whitespace, (_children, _exit_contents)))) => (
            remaining,
            vec![Object::PlainText(PlainText {
                source: whitespace.into(),
            })],
        ),
        Err(_) => {
            let (remaining, (children, _exit_contents)) =
                many_till(object_matcher, exit_matcher)(remaining)?;
            (remaining, children)
        }
    };
    let (remaining, _end) = lesser_block_end_specialized(&parser_context, remaining)?;

    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        VerseBlock {
            source: source.into(),
            name: name.into(),
            data: parameters.map(|parameters| Into::<&str>::into(parameters)),
            children,
        },
    ))
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
pub(crate) fn comment_block<'b, 'g, 'r, 's>(
    context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
) -> Res<OrgSource<'s>, CommentBlock<'s>> {
    let (remaining, name) = lesser_block_begin("comment")(context, input)?;
    let (remaining, parameters) = opt(tuple((space1, data)))(remaining)?;
    let (remaining, _nl) = recognize(tuple((space0, line_ending)))(remaining)?;
    let lesser_block_end_specialized = lesser_block_end("comment");
    let contexts = [
        ContextElement::ConsumeTrailingWhitespace(true),
        ContextElement::Context("lesser block"),
        ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &lesser_block_end_specialized,
        }),
    ];
    let parser_context = context.with_additional_node(&contexts[0]);
    let parser_context = parser_context.with_additional_node(&contexts[1]);
    let parser_context = parser_context.with_additional_node(&contexts[2]);
    let parameters = match parameters {
        Some((_ws, parameters)) => Some(parameters),
        None => None,
    };

    let (remaining, contents) = parser_with_context!(text_until_exit)(&parser_context)(remaining)?;
    let (remaining, _end) = lesser_block_end_specialized(&parser_context, remaining)?;

    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        CommentBlock {
            source: source.into(),
            name: name.into(),
            data: parameters.map(|parameters| Into::<&str>::into(parameters)),
            contents: contents.into(),
        },
    ))
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
pub(crate) fn example_block<'b, 'g, 'r, 's>(
    context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
) -> Res<OrgSource<'s>, ExampleBlock<'s>> {
    let (remaining, _name) = lesser_block_begin("example")(context, input)?;
    let (remaining, parameters) = opt(tuple((space1, data)))(remaining)?;
    let (remaining, _nl) = recognize(tuple((space0, line_ending)))(remaining)?;
    let lesser_block_end_specialized = lesser_block_end("example");
    let contexts = [
        ContextElement::ConsumeTrailingWhitespace(true),
        ContextElement::Context("lesser block"),
        ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &lesser_block_end_specialized,
        }),
    ];
    let parser_context = context.with_additional_node(&contexts[0]);
    let parser_context = parser_context.with_additional_node(&contexts[1]);
    let parser_context = parser_context.with_additional_node(&contexts[2]);
    let parameters = match parameters {
        Some((_ws, parameters)) => Some(parameters),
        None => None,
    };

    let (remaining, contents) = parser_with_context!(text_until_exit)(&parser_context)(remaining)?;
    let (remaining, _end) = lesser_block_end_specialized(&parser_context, remaining)?;

    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        ExampleBlock {
            source: source.into(),
            name: source.into(),
            data: parameters.map(|parameters| Into::<&str>::into(parameters)),
            contents: contents.into(),
        },
    ))
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
pub(crate) fn export_block<'b, 'g, 'r, 's>(
    context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
) -> Res<OrgSource<'s>, ExportBlock<'s>> {
    let (remaining, name) = lesser_block_begin("export")(context, input)?;
    // https://orgmode.org/worg/org-syntax.html#Blocks claims that export blocks must have a single word for data but testing shows no data and multi-word data still parses as an export block.
    let (remaining, parameters) = opt(tuple((space1, data)))(remaining)?;
    let (remaining, _nl) = recognize(tuple((space0, line_ending)))(remaining)?;
    let lesser_block_end_specialized = lesser_block_end("export");
    let contexts = [
        ContextElement::ConsumeTrailingWhitespace(true),
        ContextElement::Context("lesser block"),
        ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &lesser_block_end_specialized,
        }),
    ];
    let parser_context = context.with_additional_node(&contexts[0]);
    let parser_context = parser_context.with_additional_node(&contexts[1]);
    let parser_context = parser_context.with_additional_node(&contexts[2]);
    let parameters = match parameters {
        Some((_ws, parameters)) => Some(parameters),
        None => None,
    };

    let (remaining, contents) = parser_with_context!(text_until_exit)(&parser_context)(remaining)?;
    let (remaining, _end) = lesser_block_end_specialized(&parser_context, remaining)?;

    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        ExportBlock {
            source: source.into(),
            name: name.into(),
            data: parameters.map(|parameters| Into::<&str>::into(parameters)),
            contents: contents.into(),
        },
    ))
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
pub(crate) fn src_block<'b, 'g, 'r, 's>(
    context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
) -> Res<OrgSource<'s>, SrcBlock<'s>> {
    let (remaining, name) = lesser_block_begin("src")(context, input)?;
    // https://orgmode.org/worg/org-syntax.html#Blocks claims that data is mandatory and must follow the LANGUAGE SWITCHES ARGUMENTS pattern but testing has shown that no data and incorrect data here will still parse to a src block.
    let (remaining, parameters) = opt(tuple((space1, data)))(remaining)?;
    let (remaining, _nl) = recognize(tuple((space0, line_ending)))(remaining)?;
    let lesser_block_end_specialized = lesser_block_end("src");
    let contexts = [
        ContextElement::ConsumeTrailingWhitespace(true),
        ContextElement::Context("lesser block"),
        ContextElement::ExitMatcherNode(ExitMatcherNode {
            class: ExitClass::Alpha,
            exit_matcher: &lesser_block_end_specialized,
        }),
    ];
    let parser_context = context.with_additional_node(&contexts[0]);
    let parser_context = parser_context.with_additional_node(&contexts[1]);
    let parser_context = parser_context.with_additional_node(&contexts[2]);
    let parameters = match parameters {
        Some((_ws, parameters)) => Some(parameters),
        None => None,
    };

    let (remaining, contents) = parser_with_context!(text_until_exit)(&parser_context)(remaining)?;
    let (remaining, _end) = lesser_block_end_specialized(&parser_context, remaining)?;

    let source = get_consumed(input, remaining);
    Ok((
        remaining,
        SrcBlock {
            source: source.into(),
            name: name.into(),
            data: parameters.map(|parameters| Into::<&str>::into(parameters)),
            contents: contents.into(),
        },
    ))
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn name<'s>(input: OrgSource<'s>) -> Res<OrgSource<'s>, OrgSource<'s>> {
    is_not(" \t\r\n")(input)
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn data<'s>(input: OrgSource<'s>) -> Res<OrgSource<'s>, OrgSource<'s>> {
    is_not("\r\n")(input)
}

fn lesser_block_end(current_name: &str) -> impl ContextMatcher {
    let current_name_lower = current_name.to_lowercase();
    move |context, input: OrgSource<'_>| {
        _lesser_block_end(context, input, current_name_lower.as_str())
    }
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn _lesser_block_end<'b, 'g, 'r, 's, 'c>(
    _context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
    current_name_lower: &'c str,
) -> Res<OrgSource<'s>, OrgSource<'s>> {
    start_of_line(input)?;
    let (remaining, _leading_whitespace) = space0(input)?;
    let (remaining, (_begin, _name, _ws, _ending)) = tuple((
        tag_no_case("#+end_"),
        tag_no_case(current_name_lower),
        space0,
        alt((eof, line_ending)),
    ))(remaining)?;
    let source = get_consumed(input, remaining);
    Ok((remaining, source))
}

/// Parser for the beginning of a lesser block
///
/// current_name MUST be lowercase. We do not do the conversion ourselves because it is not allowed in a const fn.
const fn lesser_block_begin<'c>(current_name: &'c str) -> impl ContextMatcher + 'c {
    // TODO: Since this is a const fn, is there ANY way to "generate" functions at compile time?
    move |context, input: OrgSource<'_>| _lesser_block_begin(context, input, current_name)
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn _lesser_block_begin<'b, 'g, 'r, 's, 'c>(
    _context: RefContext<'b, 'g, 'r, 's>,
    input: OrgSource<'s>,
    current_name_lower: &'c str,
) -> Res<OrgSource<'s>, OrgSource<'s>> {
    start_of_line(input)?;
    let (remaining, _leading_whitespace) = space0(input)?;
    let (remaining, (_begin, name)) = tuple((
        tag_no_case("#+begin_"),
        verify(name, |name: &OrgSource<'_>| {
            Into::<&str>::into(name).to_lowercase().as_str() == current_name_lower
        }),
    ))(remaining)?;
    Ok((remaining, name))
}
